@@ -1,5 +1,6 @@
 import collections.abc
 import gc
+import pickle
 import random
 import unittest
 import weakref
@@ -60,7 +61,7 @@ class KeyStr(str):
         return super().__eq__(other)
 
 
-class HaskKeyCrasher:
+class HashKeyCrasher:
 
     def __init__(self, *, error_on_hash=False, error_on_eq=False,
                  error_on_repr=False):
@@ -93,13 +94,6 @@ class BaseMapTest:
 
     Map = None
 
-    def test_init_no_args(self):
-        with self.assertRaisesRegex(TypeError, 'positional argument'):
-            self.Map(dict(a=1))
-
-        with self.assertRaisesRegex(TypeError, 'keyword argument'):
-            self.Map(a=1)
-
     def test_hashkey_helper_1(self):
         k1 = HashKey(10, 'aaa')
         k2 = HashKey(10, 'bbb')
@@ -260,14 +254,14 @@ class BaseMapTest:
                 key = KeyStr(i)
 
                 if not (i % CRASH_HASH_EVERY):
-                    with HaskKeyCrasher(error_on_hash=True):
+                    with HashKeyCrasher(error_on_hash=True):
                         with self.assertRaises(HashingError):
                             h.set(key, i)
 
                 h = h.set(key, i)
 
                 if not (i % CRASH_EQ_EVERY):
-                    with HaskKeyCrasher(error_on_eq=True):
+                    with HashKeyCrasher(error_on_eq=True):
                         with self.assertRaises(EqError):
                             h.get(KeyStr(i))  # really trigger __eq__
 
@@ -289,12 +283,12 @@ class BaseMapTest:
                 key = KeyStr(i)
 
                 if not (iter_i % CRASH_HASH_EVERY):
-                    with HaskKeyCrasher(error_on_hash=True):
+                    with HashKeyCrasher(error_on_hash=True):
                         with self.assertRaises(HashingError):
                             h.delete(key)
 
                 if not (iter_i % CRASH_EQ_EVERY):
-                    with HaskKeyCrasher(error_on_eq=True):
+                    with HashKeyCrasher(error_on_eq=True):
                         with self.assertRaises(EqError):
                             h.delete(KeyStr(i))
 
@@ -644,6 +638,17 @@ class BaseMapTest:
             set(list(it)),
             {(A, 'a'), (B, 'b'), (C, 'c'), (D, 'd'), (E, 'e'), (F, 'f')})
 
+    def test_map_items_3(self):
+        h = self.Map()
+        self.assertEqual(len(h.items()), 0)
+        self.assertEqual(list(h.items()), [])
+
+    def test_map_items_4(self):
+        h = self.Map(a=1, b=2, c=3)
+        k = h.items()
+        self.assertEqual(set(k), {('a', 1), ('b', 2), ('c', 3)})
+        self.assertEqual(set(k), {('a', 1), ('b', 2), ('c', 3)})
+
     def test_map_keys_1(self):
         A = HashKey(100, 'A')
         B = HashKey(101, 'B')
@@ -663,6 +668,12 @@ class BaseMapTest:
         self.assertEqual(set(list(h.keys())), {A, B, C, D, E, F})
         self.assertEqual(set(list(h)), {A, B, C, D, E, F})
 
+    def test_map_keys_2(self):
+        h = self.Map(a=1, b=2, c=3)
+        k = h.keys()
+        self.assertEqual(set(k), {'a', 'b', 'c'})
+        self.assertEqual(set(k), {'a', 'b', 'c'})
+
     def test_map_values_1(self):
         A = HashKey(100, 'A')
         B = HashKey(101, 'B')
@@ -681,10 +692,11 @@ class BaseMapTest:
 
         self.assertEqual(set(list(h.values())), {'a', 'b', 'c', 'd', 'e', 'f'})
 
-    def test_map_items_3(self):
-        h = self.Map()
-        self.assertEqual(len(h.items()), 0)
-        self.assertEqual(list(h.items()), [])
+    def test_map_values_2(self):
+        h = self.Map(a=1, b=2, c=3)
+        k = h.values()
+        self.assertEqual(set(k), {1, 2, 3})
+        self.assertEqual(set(k), {1, 2, 3})
 
     def test_map_eq_1(self):
         A = HashKey(100, 'A')
@@ -783,7 +795,7 @@ class BaseMapTest:
         h = h.set(A, h)
 
         ref = weakref.ref(h)
-        hi = h.items()
+        hi = iter(h.items())
         next(hi)
 
         del h, hi
@@ -807,11 +819,11 @@ class BaseMapTest:
         self.assertFalse(B in h)
 
         with self.assertRaises(EqError):
-            with HaskKeyCrasher(error_on_eq=True):
+            with HashKeyCrasher(error_on_eq=True):
                 AA in h
 
         with self.assertRaises(HashingError):
-            with HaskKeyCrasher(error_on_hash=True):
+            with HashKeyCrasher(error_on_hash=True):
                 AA in h
 
     def test_map_getitem_1(self):
@@ -830,11 +842,11 @@ class BaseMapTest:
             h[B]
 
         with self.assertRaises(EqError):
-            with HaskKeyCrasher(error_on_eq=True):
+            with HashKeyCrasher(error_on_eq=True):
                 h[AA]
 
         with self.assertRaises(HashingError):
-            with HaskKeyCrasher(error_on_hash=True):
+            with HashKeyCrasher(error_on_hash=True):
                 h[AA]
 
     def test_repr_1(self):
@@ -850,11 +862,11 @@ class BaseMapTest:
         A = HashKey(100, 'A')
 
         with self.assertRaises(ReprError):
-            with HaskKeyCrasher(error_on_repr=True):
+            with HashKeyCrasher(error_on_repr=True):
                 repr(h.set(1, 2).set(A, 3).set(3, 4))
 
         with self.assertRaises(ReprError):
-            with HaskKeyCrasher(error_on_repr=True):
+            with HashKeyCrasher(error_on_repr=True):
                 repr(h.set(1, 2).set(2, A).set(3, 4))
 
     def test_repr_3(self):
@@ -895,17 +907,356 @@ class BaseMapTest:
 
         m = h.set(1, 2).set(A, 3).set(3, 4)
         with self.assertRaises(HashingError):
-            with HaskKeyCrasher(error_on_hash=True):
+            with HashKeyCrasher(error_on_hash=True):
                 hash(m)
 
         m = h.set(1, 2).set(2, A).set(3, 4)
         with self.assertRaises(HashingError):
-            with HaskKeyCrasher(error_on_hash=True):
+            with HashKeyCrasher(error_on_hash=True):
                 hash(m)
 
     def test_abc_1(self):
         self.assertTrue(issubclass(self.Map, collections.abc.Mapping))
 
+    def test_map_mut_1(self):
+        h = self.Map()
+        h = h.set('a', 1)
+
+        hm1 = h.mutate()
+        hm2 = h.mutate()
+
+        self.assertFalse(isinstance(hm1, self.Map))
+
+        self.assertIsNot(hm1, hm2)
+        self.assertEqual(hm1['a'], 1)
+        self.assertEqual(hm2['a'], 1)
+
+        hm1.set('b', 2)
+        hm1.set('c', 3)
+
+        hm2.set('x', 100)
+        hm2.set('a', 1000)
+
+        self.assertEqual(hm1['a'], 1)
+        self.assertEqual(hm1.get('x', -1), -1)
+
+        self.assertEqual(hm2['a'], 1000)
+        self.assertTrue('x' in hm2)
+
+        h1 = hm1.finish()
+        h2 = hm2.finish()
+
+        self.assertTrue(isinstance(h1, self.Map))
+
+        self.assertEqual(dict(h.items()), {'a': 1})
+        self.assertEqual(dict(h1.items()), {'a': 1, 'b': 2, 'c': 3})
+        self.assertEqual(dict(h2.items()), {'a': 1000, 'x': 100})
+
+    def test_map_mut_2(self):
+        h = self.Map()
+        h = h.set('a', 1)
+
+        hm1 = h.mutate()
+        hm1.set('a', 2)
+        hm1.set('a', 3)
+        hm1.set('a', 4)
+        h2 = hm1.finish()
+
+        self.assertEqual(dict(h.items()), {'a': 1})
+        self.assertEqual(dict(h2.items()), {'a': 4})
+
+    def test_map_mut_3(self):
+        h = self.Map()
+        h = h.set('a', 1)
+        hm1 = h.mutate()
+
+        self.assertTrue(repr(hm1).startswith(
+            "<immutables.MapMutation({'a': 1})"))
+
+        with self.assertRaisesRegex(TypeError, 'unhashable type'):
+            hash(hm1)
+
+    def test_map_mut_4(self):
+        h = self.Map()
+        h = h.set('a', 1)
+        h = h.set('b', 2)
+
+        hm1 = h.mutate()
+        hm2 = h.mutate()
+
+        self.assertEqual(hm1, hm2)
+
+        hm1.set('a', 10)
+        self.assertNotEqual(hm1, hm2)
+
+        hm2.set('a', 10)
+        self.assertEqual(hm1, hm2)
+
+        self.assertEqual(hm2.pop('a'), 10)
+        self.assertNotEqual(hm1, hm2)
+
+    def test_map_mut_5(self):
+        h = self.Map({'a': 1, 'b': 2}, z=100)
+        self.assertTrue(isinstance(h, self.Map))
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+
+        h2 = h.update(z=200, y=-1)
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+        self.assertEqual(dict(h2.items()), {'a': 1, 'b': 2, 'z': 200, 'y': -1})
+
+        h3 = h2.update([(1, 2), (3, 4)])
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+        self.assertEqual(dict(h2.items()), {'a': 1, 'b': 2, 'z': 200, 'y': -1})
+        self.assertEqual(dict(h3.items()),
+                         {'a': 1, 'b': 2, 'z': 200, 'y': -1, 1: 2, 3: 4})
+
+        h4 = h3.update()
+        self.assertIs(h4, h3)
+
+        h5 = h4.update(self.Map({'zzz': 'yyz'}))
+
+        self.assertEqual(dict(h5.items()),
+                         {'a': 1, 'b': 2, 'z': 200, 'y': -1, 1: 2, 3: 4,
+                          'zzz': 'yyz'})
+
+    def test_map_mut_6(self):
+        h = self.Map({'a': 1, 'b': 2}, z=100)
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+
+        with self.assertRaisesRegex(TypeError, 'not iterable'):
+            h.update(1)
+
+        with self.assertRaisesRegex(ValueError, 'map update sequence element'):
+            h.update([(1, 2), (3, 4, 5)])
+
+        with self.assertRaisesRegex(TypeError, 'cannot convert map update'):
+            h.update([(1, 2), 1])
+
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+
+    def test_map_mut_7(self):
+        key = HashKey(123, 'aaa')
+
+        h = self.Map({'a': 1, 'b': 2}, z=100)
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+
+        upd = {key: 1}
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                h.update(upd)
+
+        upd = self.Map({key: 'zzz'})
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                h.update(upd)
+
+        upd = [(1, 2), (key, 'zzz')]
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                h.update(upd)
+
+        self.assertEqual(dict(h.items()), {'a': 1, 'b': 2, 'z': 100})
+
+    def test_map_mut_8(self):
+        key1 = HashKey(123, 'aaa')
+        key2 = HashKey(123, 'bbb')
+
+        h = self.Map({key1: 123})
+        self.assertEqual(dict(h.items()), {key1: 123})
+
+        upd = {key2: 1}
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                h.update(upd)
+
+        upd = self.Map({key2: 'zzz'})
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                h.update(upd)
+
+        upd = [(1, 2), (key2, 'zzz')]
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                h.update(upd)
+
+        self.assertEqual(dict(h.items()), {key1: 123})
+
+    def test_map_mut_9(self):
+        key1 = HashKey(123, 'aaa')
+
+        src = {key1: 123}
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                self.Map(src)
+
+        src = self.Map({key1: 123})
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                self.Map(src)
+
+        src = [(1, 2), (key1, 123)]
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                self.Map(src)
+
+    def test_map_mut_10(self):
+        key1 = HashKey(123, 'aaa')
+
+        m = self.Map({key1: 123})
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                del mm[key1]
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                mm.pop(key1, None)
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_hash=True):
+            with self.assertRaises(HashingError):
+                mm.set(key1, 123)
+
+    def test_map_mut_11(self):
+        m = self.Map({'a': 1, 'b': 2})
+
+        mm = m.mutate()
+        self.assertEqual(mm.pop('a', 1), 1)
+        self.assertEqual(mm.finish(), self.Map({'b': 2}))
+
+        mm = m.mutate()
+        self.assertEqual(mm.pop('b', 1), 2)
+        self.assertEqual(mm.finish(), self.Map({'a': 1}))
+
+        mm = m.mutate()
+        self.assertEqual(mm.pop('b', 1), 2)
+        del mm['a']
+        self.assertEqual(mm.finish(), self.Map())
+
+    def test_map_mut_12(self):
+        m = self.Map({'a': 1, 'b': 2})
+
+        mm = m.mutate()
+        mm.finish()
+
+        with self.assertRaisesRegex(ValueError, 'has been finished'):
+            mm.pop('a')
+
+        with self.assertRaisesRegex(ValueError, 'has been finished'):
+            del mm['a']
+
+        with self.assertRaisesRegex(ValueError, 'has been finished'):
+            mm.set('a', 'b')
+
+        with self.assertRaisesRegex(ValueError, 'has been finished'):
+            mm['a'] = 'b'
+
+    def test_map_mut_13(self):
+        key1 = HashKey(123, 'aaa')
+        key2 = HashKey(123, 'aaa')
+
+        m = self.Map({key1: 123})
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                del mm[key2]
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                mm.pop(key2, None)
+
+        mm = m.mutate()
+        with HashKeyCrasher(error_on_eq=True):
+            with self.assertRaises(EqError):
+                mm.set(key2, 123)
+
+    def test_map_mut_14(self):
+        m = self.Map(a=1, b=2)
+
+        with m.mutate() as mm:
+            mm['z'] = 100
+            del mm['a']
+
+        self.assertEqual(mm.finish(), self.Map(z=100, b=2))
+
+    def test_map_mut_15(self):
+        m = self.Map(a=1, b=2)
+
+        with self.assertRaises(ZeroDivisionError):
+            with m.mutate() as mm:
+                mm['z'] = 100
+                del mm['a']
+                1 / 0
+
+        self.assertEqual(mm.finish(), self.Map(z=100, b=2))
+        self.assertEqual(m, self.Map(a=1, b=2))
+
+    def test_map_mut_stress(self):
+        COLLECTION_SIZE = 7000
+        TEST_ITERS_EVERY = 647
+        RUN_XTIMES = 3
+
+        for _ in range(RUN_XTIMES):
+            h = self.Map()
+            d = dict()
+
+            for i in range(COLLECTION_SIZE // TEST_ITERS_EVERY):
+
+                hm = h.mutate()
+                for j in range(TEST_ITERS_EVERY):
+                    key = random.randint(1, 100000)
+                    key = HashKey(key % 271, str(key))
+
+                    hm.set(key, key)
+                    d[key] = key
+
+                    self.assertEqual(len(hm), len(d))
+
+                h2 = hm.finish()
+                self.assertEqual(dict(h2.items()), d)
+                h = h2
+
+            self.assertEqual(dict(h.items()), d)
+            self.assertEqual(len(h), len(d))
+
+            it = iter(tuple(d.keys()))
+            for i in range(COLLECTION_SIZE // TEST_ITERS_EVERY):
+
+                hm = h.mutate()
+                for j in range(TEST_ITERS_EVERY):
+                    try:
+                        key = next(it)
+                    except StopIteration:
+                        break
+
+                    del d[key]
+                    del hm[key]
+
+                    self.assertEqual(len(hm), len(d))
+
+                h2 = hm.finish()
+                self.assertEqual(dict(h2.items()), d)
+                h = h2
+
+            self.assertEqual(dict(h.items()), d)
+            self.assertEqual(len(h), len(d))
+
+    def test_map_pickle(self):
+        h = self.Map(a=1, b=2)
+        for proto in range(pickle.HIGHEST_PROTOCOL):
+            p = pickle.dumps(h, proto)
+            uh = pickle.loads(p)
+
+            self.assertTrue(isinstance(uh, self.Map))
+            self.assertEqual(h, uh)
+
+        with self.assertRaisesRegex(TypeError, "can't pickle"):
+            pickle.dumps(h.mutate())
+
 
 class PyMapTest(BaseMapTest, unittest.TestCase):
 
